import inspect
import logging

from ..task.base import compile_scheduler_options_v2
from ..task.base.base_task import schedulers

logger = logging.getLogger()


_scheduler_presets = {
    'auto': {'scheduler': 'local', 'searcher': 'bayesopt'},
    'random': {'scheduler': 'local', 'searcher': 'random'},
    'bayesopt': {'scheduler': 'local', 'searcher': 'bayesopt'},
    # Don't include hyperband and bayesopt hyperband at present
}


def scheduler_factory(
        hyperparameter_tune_kwargs,
        time_out: float = None,
        num_trials: int = None,
        nthreads_per_trial='all',
        ngpus_per_trial='all',
        **kwargs):
    """
    Constructs a scheduler via lazy initialization based on the input hyperparameter_tune_kwargs.
    The output will contain the scheduler class and init arguments except for the `train_fn` argument, which must be specified downstream.

    Parameters
    ----------
    hyperparameter_tune_kwargs : str or dict
        Hyperparameter tuning strategy and kwargs.
        If None, then hyperparameter tuning will not be performed.
        Valid preset values:
            'auto': Uses the 'bayesopt' preset.
            'random': Performs HPO via random search using local scheduler.
            'bayesopt': Performs HPO via bayesian optimization using local scheduler.
        For valid dictionary keys, refer to :class:`autogluon.core.scheduler.FIFOScheduler` documentation.
            The 'searcher' key is required when providing a dict.
            Some schedulers may have different valid keys.
    time_out : float, default = None
        Same as hyperparameter_tune_kwargs['time_out']. Ignored if specified in hyperparameter_tune_kwargs.
        At least one of time_out or num_trials must be specified.
    num_trials : int, default = None
        Same as hyperparameter_tune_kwargs['time_out']. Ignored if specified in hyperparameter_tune_kwargs.
        At least one of time_out or num_trials must be specified.
    nthreads_per_trial : str or int, default = 'all'
        Same as hyperparameter_tune_kwargs['nthreads_per_trial']. Ignored if specified in hyperparameter_tune_kwargs.
        Number of CPU threads to use per HPO trial.
        Valid str values:
            'all': Use all CPU threads.
            'auto': Keep value as 'auto' in output, must be updated downstream.
        If None, use all CPU threads as in 'all'.
    ngpus_per_trial : str or int, default = 'all'
        Same as hyperparameter_tune_kwargs['ngpus_per_trial']. Ignored if specified in hyperparameter_tune_kwargs.
        Number of GPUs to use per HPO trial.
        Valid str values:
            'all': Use all GPUs.
            'auto': Keep value as 'auto' in output, must be updated downstream.
        If None, use 0 GPUs.
    **kwargs :
        Kwargs to specify any other scheduler parameters.
        A kwarg will be ignored if also specified in hyperparameter_tune_kwargs.

    Returns
    -------
    scheduler_cls : class :class:`autogluon.core.scheduler.TaskScheduler`, scheduler_params : dict
        scheduler_cls is the class of the scheduler that will be constructed.
        scheduler_params is the key word parameter arguments to pass to the Scheeduler class constructor when initializing a Scheduler object.
        To actually construct a Scheduler object, call `scheduler_cls(train_fn, **scheduler_params)`
        By default in scheduler_params: time_attr='epoch', reward_attr='validation_performance'

    Examples
    --------
    >>> import numpy as np
    >>> import autogluon.core as ag
    >>> from autogluon.core.scheduler.scheduler_factory import scheduler_factory
    >>>
    >>>
    >>> @ag.args()
    >>> def train_fn(args, reporter):
    >>>     for e in range(args.epochs):
    >>>         dummy_accuracy = 1 - np.power(1.8, -np.random.uniform(e, 2*e))
    >>>         reporter(epoch=e+1, validation_performance=dummy_accuracy, lr=args.lr, wd=args.wd)
    >>>
    >>> hyperparameters = dict(
    >>>     lr=ag.space.Real(1e-3, 1e-2, log=True),
    >>>     wd=ag.space.Real(1e-3, 1e-2),
    >>>     epochs=10,
    >>> )
    >>>
    >>> scheduler_cls, scheduler_params = scheduler_factory('auto', num_trials=5)
    >>>
    >>> train_fn.register_args(**hyperparameters)  # Register search space
    >>> scheduler = scheduler_cls(train_fn, **scheduler_params)
    >>>
    >>> scheduler.run()
    >>> scheduler.join_jobs()
    >>> scheduler.get_training_curves(plot=True)
    """
    if hyperparameter_tune_kwargs is None:
        raise ValueError(f"hyperparameter_tune_kwargs cannot be None.")
    if isinstance(hyperparameter_tune_kwargs, str):
        hyperparameter_tune_kwargs = get_hyperparameter_tune_kwargs_preset(hyperparameter_tune_kwargs)
    if not isinstance(hyperparameter_tune_kwargs, dict):
        raise ValueError(f"hyperparameter_tune_kwargs must be of type str or dict, but is type: {type(hyperparameter_tune_kwargs)}")
    if 'scheduler' not in hyperparameter_tune_kwargs:
        raise ValueError(f"Required key 'scheduler' is not present in hyperparameter_tune_kwargs: {hyperparameter_tune_kwargs}")
    if 'searcher' not in hyperparameter_tune_kwargs:
        raise ValueError(f"Required key 'searcher' is not present in hyperparameter_tune_kwargs: {hyperparameter_tune_kwargs}")
    if num_trials is None and time_out is not None:
        num_trials = 1000

    scheduler_params = compile_scheduler_options_v2(
        scheduler_options=hyperparameter_tune_kwargs,
        nthreads_per_trial=nthreads_per_trial,
        ngpus_per_trial=ngpus_per_trial,
        num_trials=num_trials,
        time_out=time_out,
        **kwargs,
    )

    scheduler_cls = hyperparameter_tune_kwargs.get('scheduler', 'unknown')
    if isinstance(scheduler_cls, str):
        scheduler_cls = get_scheduler_from_preset(scheduler_cls)
    if not inspect.isclass(scheduler_cls):
        raise ValueError(f'scheduler_cls must be a class, but was instead: {scheduler_cls}')

    if scheduler_params['time_out'] is None:
        scheduler_params.pop('time_out', None)
    return scheduler_cls, scheduler_params


def get_scheduler_from_preset(scheduler_cls):
    scheduler_cls = scheduler_cls.lower()
    if scheduler_cls not in schedulers.keys():
        raise ValueError(f"Required key 'scheduler' in hyperparameter_tune_kwargs must be one of the "
                         f"values {schedulers.keys()}, but was instead: {scheduler_cls}")
    scheduler_cls = schedulers.get(scheduler_cls)
    return scheduler_cls


def get_hyperparameter_tune_kwargs_preset(preset: str):
    if preset not in _scheduler_presets:
        raise ValueError(f'Invalid hyperparameter_tune_kwargs preset value "{preset}". Valid presets: {list(_scheduler_presets.keys())}')
    return _scheduler_presets[preset].copy()



import os
import time
import signal
import atexit
import weakref
import logging
import subprocess
import concurrent
from threading import Thread 
import multiprocessing as mp
from distributed import Client

from .ssh_helper import start_scheduler, start_worker

__all__ = ['Remote']

logger = logging.getLogger(__name__)

_global_remote_services = weakref.WeakValueDictionary()
_global_service_index = [0]

def _get_global_remote_service():
    L = sorted(list(_global_remote_services), reverse=True)
    for k in L:
        c = _global_remote_services[k]
        if c.status != "closed":
            return c
        else:
            del _global_remote_services[k]
    del L
    return None

def _set_global_remote_service(c):
    if c is not None:
        _global_remote_services[_global_service_index[0]] = c
        _global_service_index[0] += 1

def _close_global_remote_services():
    """
    Force close of global client.  This cleans up when a client
    wasn't close explicitly, e.g. interactive sessions.
    """
    c = _get_global_remote_service()
    if c is not None:
        c.shutdown()

class Service(object):
    def __init__(self, proc):
        self.proc = proc
        self.status = 'live'

    def shutdown(self):
        os.killpg(os.getpgid(self.proc.pid), signal.SIGTERM)

def start_service(remote_ip, port):
    cmd = ['agremote', '--address', remote_ip, '--port', str(port)]
    proc = subprocess.Popen(cmd)
    return Service(proc)

class Remote(Client):
    LOCK = mp.Lock()
    REMOTE_ID = mp.Value('i', 0)
    def __init__(self, remote_ip=None, port=None, local=False, ssh_username=None,
            ssh_port=22, ssh_private_key=None, remote_python=None, timeout=60):
        self.service = None
        if local:
            super().__init__(processes=False)
        else:
            remote_addr = (remote_ip + ':{}'.format(port))
            self.service = start_service(remote_ip, port)
            _set_global_remote_service(self.service)
            super().__init__(remote_addr, timeout=timeout)
        with Remote.LOCK:
            self.remote_id = Remote.REMOTE_ID.value
            Remote.REMOTE_ID.value += 1

    def close(self, timeout=2):
        if self.service:
            self.service.shutdown()
        super().close(timeout)
        self.shutdown()

    def upload_files(self, files, **kwargs):
        for filename in files:
            self.upload_file(filename, **kwargs)

    def __repr__(self):
        reprstr = self.__class__.__name__ + ' REMOTE_ID: {}, \n\t'.format(self.remote_id) + \
            super().__repr__()
        return reprstr

class DaskRemoteService(object):
    def __init__(self, remote_addr, scheduler_port, ssh_username=None,
        ssh_port=22, ssh_private_key=None, remote_python=None):

        self.scheduler_addr = remote_addr
        self.scheduler_port = scheduler_port

        self.ssh_username = ssh_username
        self.ssh_port = ssh_port
        self.ssh_private_key = ssh_private_key
        self.remote_python = remote_python
        self.monitor_thread = Thread()

        # Start the scheduler node
        self.scheduler = start_scheduler(
            remote_addr,
            scheduler_port,
            ssh_username,
            ssh_port,
            ssh_private_key,
            remote_python,
        )
        # Start worker nodes
        self.worker = start_worker(
            self.scheduler_addr,
            self.scheduler_port,
            remote_addr,
            self.ssh_username,
            self.ssh_port,
            self.ssh_private_key,
            self.remote_python,
        )
        self.start_monitoring()
        self.status = "live"

    def start_monitoring(self):
        if self.monitor_thread.is_alive():
            return
        self.monitor_thread = Thread(target=self.monitor_remote_processes)
        #self.monitor_thread.daemon = True
        self.monitor_thread.start()

    def monitor_remote_processes(self):
        all_processes = [self.scheduler, self.worker]
        try:
            while True:
                for process in all_processes:
                    while not process["output_queue"].empty():
                        try:
                            msg = process["output_queue"].get()
                            if 'distributed.' in msg:
                                msg = msg.replace('distributed', 'autogluon')

                            print(msg)
                        except Exception as e:
                            print(f'Exception happend {e}, terminating the remote.')
                            break
                # Kill some time and free up CPU
                time.sleep(0.1)

        except KeyboardInterrupt:
            pass

    def shutdown(self):
        all_processes = [self.worker, self.scheduler]

        for process in all_processes:
            process["input_queue"].put("shutdown")
            process["thread"].join()
        self.status = "closed"

atexit.register(_close_global_remote_services)



import numpy as np
from typing import Callable, Type, Tuple, Set, Iterable, Optional, Any
import ConfigSpace as CS
from collections import Counter
import logging

from .gp_fifo_searcher import GET_CONFIG_RANDOM_RETRIES, GPSearcher, MapReward, \
    create_initial_candidates_scorer, decode_state
from ..datatypes.common import PendingEvaluation, candidate_for_print
from ..datatypes.config_ext import ExtendedConfiguration
from ..datatypes.hp_ranges_cs import HyperparameterRanges_CS
from ..datatypes.tuning_job_state import TuningJobState
from ..models.gp_model import GPModel
from ..models.gpmodel_skipopt import SkipOptimizationPredicate
from ..models.gpmodel_transformers import GPModelArgs
from ..tuning_algorithms.base_classes import LocalOptimizer, \
    AcquisitionFunction, DEFAULT_METRIC
from ..tuning_algorithms.bo_algorithm import BayesianOptimizationAlgorithm
from ..tuning_algorithms.common import compute_blacklisted_candidates
from ..tuning_algorithms.defaults import DEFAULT_LOCAL_OPTIMIZER_CLASS, DEFAULT_NUM_INITIAL_CANDIDATES, \
    DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS
from ..utils.debug_log import DebugLogPrinter
from ..utils.duplicate_detector import DuplicateDetectorIdentical
from ..utils.simple_profiler import SimpleProfiler

logger = logging.getLogger(__name__)


class GPMultiFidelitySearcher(GPSearcher):
    """
    Supports asynchronous multi-fidelity hyperparameter optimization, in the
    style of Hyperband or BOHB. Here, a joint GP surrogate model is fit to
    observations made at all resource levels.

    """
    def __init__(
            self, hp_ranges: HyperparameterRanges_CS,
            resource_attr_key: str,
            resource_attr_range: Tuple[int, int],
            random_seed: int,
            gpmodel: GPModel, model_args: GPModelArgs,
            map_reward: MapReward,
            acquisition_class: Type[AcquisitionFunction],
            resource_for_acquisition: Callable[..., int],
            init_state: TuningJobState = None,
            local_minimizer_class: Type[LocalOptimizer] = DEFAULT_LOCAL_OPTIMIZER_CLASS,
            skip_optimization: SkipOptimizationPredicate = None,
            num_initial_candidates: int = DEFAULT_NUM_INITIAL_CANDIDATES,
            num_initial_random_choices: int = DEFAULT_NUM_INITIAL_RANDOM_EVALUATIONS,
            initial_scoring: Optional[str] = None,
            first_is_default: bool = True,
            debug_log: Optional[DebugLogPrinter] = None,
            cost_metric_name: Optional[str] = None,
            profiler: Optional[SimpleProfiler] = None,
            getconfig_callback: Optional[Callable[[dict], Any]] = None):
        """
        Note that the GPMXNetModel is created on demand (by the state
        transformer) in get_config, along with components needed for the BO
        algorithm.

        The configuration space is hp_ranges. This does not include the resource
        attribute, which is passed as result component instead, with key
        resource_attr_key. The GP model is over configuration and resource
        attribute, its configuration space is maintained in configspace_ext.

        The search for a next candidate in get_config is fixing the resource
        level, meaning that extended configs for which the acquisition function
        is evaluated, all have the same resource level. This level may depend on
        the current state, the function can be passed as
        resource_for_acquisition. Its signature is
            resource_for_acquisition(state, resource_attr_name, **kwargs) -> int,
        where state is TuningJobState.
        Example: resource_for_acquisition may count the number of labeled data
        at each resource level, and choose the largest level supported by
        enough data. Or information about bracket and first milestone may be
        passed from the scheduler via **kwargs.

        The searcher is supposed to maximize reward, while internally, the
        criterion is minimized. map_reward maps reward to internal criterion, it
        must be strictly decreasing.

        :param hp_ranges: Configuration space without resource attribute
        :param resource_attr_key: Key for resource attribute.
            Note: The resource attribute must be int valued
        :param resource_attr_range: Range (lower, upper) for resource
            attribute( (must be int valued)
        :param random_seed:
        :param gpmodel: GP regression model
        :param model_args: Arguments for GPMXNet model creation
        :param map_reward: Function mapping reward to criterion to be minimized
        :param acquisition_class: Type for acquisition function
        :param resource_for_acquisition: See above
        :param init_state: TuningJobState to start from (default is empty).
            Here, init_state.hp_ranges must be equal to
            self.configspace_ext.hp_ranges_ext here (this is not checked)
        :param local_minimizer_class: Type for local minimizer
        :param skip_optimization: Predicate, see
            GPMXNetPendingCandidateStateTransformer
        :param num_initial_candidates: See BayesianOptimizationAlgorithm
        :param num_initial_random_choices: Configs are sampled at random until
            this many candidates received label feedback
        :param initial_scoring: Scoring function to rank initial candidates.
            Default: thompson_indep (independent Thompson sampling)
        :param first_is_default: If true, the first result of get_config is the
            default config of hp_ranges
        :param debug_log: DebugLogPrinter for debug logging (optional)
        :param cost_metric_name: If an `elapsed_time` record is passed to
            `update`, the value is entered in the state using this key
        :param profiler: If given, HPO computations are profiled
        :param getconfig_callback: If given, this callback function is called
            at the end of each get_config. It receives the chosen config
            (as dict) as single argument. One use case for this callback is
            to store profiler records.

        """
        assert isinstance(hp_ranges, HyperparameterRanges_CS), \
            "hp_ranges must be of type HyperparameterRanges_CS"
        # Extended configuration space including resource attribute
        self.configspace_ext = ExtendedConfiguration(
            hp_ranges, resource_attr_key, resource_attr_range)
        super().__init__(
            hp_ranges, random_seed, gpmodel, model_args, map_reward,
            acquisition_class, init_state, local_minimizer_class,
            skip_optimization, num_initial_candidates,
            num_initial_random_choices, initial_scoring, first_is_default,
            debug_log, cost_metric_name, profiler, getconfig_callback)

        self.resource_for_acquisition = resource_for_acquisition
        self._gpmodel = gpmodel
        if debug_log is not None:
            # Configure DebugLogPrinter
            debug_log.set_configspace_ext(self.configspace_ext)

    def _hp_ranges_ext(self):
        return self.configspace_ext.hp_ranges_ext

    def _config_ext_update(self, config, **kwargs):
        return self.configspace_ext.get(config, kwargs['resource'])

    def register_pending(self, config: CS.Configuration, milestone: int):
        """
        Registers config as pending for resource level milestone. This means
        the corresponding evaluation task is running and should reach that
        level later, when update is called for it.

        :param config:
        :param milestone:
        """
        # It is OK for the candidate already to be registered as pending, in
        # which case we do nothing
        state = self.state_transformer.state
        config_ext = self.configspace_ext.get(config, milestone)
        if config_ext not in state.pending_candidates:
            if config_ext in (x.candidate for x in state.candidate_evaluations):
                evals = state.candidate_evaluations
                num_labeled = len(evals)
                pos_cand = next(
                    i for i, x in enumerate(evals) if x.candidate == config_ext)
                error_msg = """
                This configuration is already registered as labeled:
                   Position of labeled candidate: {} of {}
                   Label value: {}
                   Resource level: {}
                """.format(
                    pos_cand, num_labeled,
                    evals[pos_cand].metrics[DEFAULT_METRIC], milestone)
                assert False, error_msg
            self.state_transformer.append_candidate(config_ext)

    def _get_unique_candidates(
            self, candidate_list: Iterable[CS.Configuration],
            target_resource: int) -> Set[CS.Configuration]:
        return set(self.configspace_ext.remap_resource(x, target_resource)
                   for x in candidate_list)

    def _get_blacklisted_candidates(self, target_resource: int) -> Set[CS.Configuration]:
        """
        We want to blacklist all configurations which are labeled or pending at
        any resource level. As the search affected by the blacklist happens at
        resource level target_resource, the candidates have to be modified
        accordingly

        """
        return self._get_unique_candidates(
            compute_blacklisted_candidates(self.state_transformer.state),
            target_resource)

    def _fix_resource_attribute(self, resource_attr_value: int):
        self.configspace_ext.hp_ranges_ext.value_for_last_pos = \
            resource_attr_value

    def get_config(self, **kwargs) -> CS.Configuration:
        """
        Runs Bayesian optimization in order to suggest the next config to evaluate.

        :return: Next config to evaluate at

        """
        state = self.state_transformer.state
        if self.do_profile:
            # Start new profiler block
            meta = {
                'fit_hyperparams': not self.state_transformer.skip_optimization(
                    state),
                'num_observed': len(state.candidate_evaluations),
                'num_pending': len(state.pending_evaluations)
            }
            self.profiler.begin_block(meta)
            self.profiler.start('all')
        # Fix resource attribute during the search to the value obtained from
        # self.resource_for_acquisition. Compute blacklisted_candidates.
        if state.candidate_evaluations:
            target_resource = self.resource_for_acquisition(
                state, self.configspace_ext.resource_attr_name, **kwargs)
        else:
            # Any valid value works here:
            target_resource = self.configspace_ext.resource_attr_range[0]
        blacklisted_candidates = self._get_blacklisted_candidates(target_resource)
        pick_random = (len(blacklisted_candidates) < self.num_initial_random_choices) or \
            (not state.candidate_evaluations)
        if self.debug_log is not None:
            self.debug_log.start_get_config('random' if pick_random else 'BO')
        if pick_random:
            config = None
            if self.first_is_default and (not blacklisted_candidates):
                # Use default configuration if there is one specified
                default_config = self.hp_ranges.config_space.get_default_configuration()
                if default_config and len(default_config.get_dictionary()) > 0:
                    config = default_config
                    if self.debug_log is not None:
                        logger.info("Start with default config:\n{}".format(
                            candidate_for_print(config)))
            if config is None:
                if self.do_profile:
                    self.profiler.start('random')
                config, _ = draw_random_candidate(
                    blacklisted_candidates, self.configspace_ext,
                    self.random_state, target_resource)
                if self.do_profile:
                    self.profiler.stop('random')
        else:
            # Obtain current SurrogateModel from state transformer. Based on
            # this, the BO algorithm components can be constructed
            state = self.state_transformer.state
            if self.do_profile:
                self.profiler.push_prefix('getconfig')
                self.profiler.start('all')
                self.profiler.start('gpmodel')
            # Note: Asking for the model triggers the posterior computation
            model = self.state_transformer.model()
            if self.do_profile:
                self.profiler.stop('gpmodel')
            # BO should only search over configs at resource level
            # target_resource
            self._fix_resource_attribute(target_resource)
            if self.debug_log is not None:
                self.debug_log.append_extra(
                    "Score values computed at target_resource = {}".format(
                        target_resource))
            # Create BO algorithm
            initial_candidates_scorer = create_initial_candidates_scorer(
                self.initial_scoring, model, self.acquisition_class,
                self.random_state)
            local_optimizer = self.local_minimizer_class(
                state, model, self.acquisition_class)
            bo_algorithm = BayesianOptimizationAlgorithm(
                initial_candidates_generator=self.random_generator,
                initial_candidates_scorer=initial_candidates_scorer,
                num_initial_candidates=self.num_initial_candidates,
                local_optimizer=local_optimizer,
                pending_candidate_state_transformer=None,
                blacklisted_candidates=blacklisted_candidates,
                num_requested_candidates=1,
                greedy_batch_selection=False,
                duplicate_detector=DuplicateDetectorIdentical(),
                profiler=self.profiler,
                sample_unique_candidates=False,
                debug_log=self.debug_log)
            # Next candidate decision
            _config = bo_algorithm.next_candidates()
            assert len(_config) > 0, \
                ("Failed to find a configuration not already chosen "
                 "before. Maybe there are no free configurations left? "
                 "The blacklist size is {}".format(len(blacklisted_candidates)))
            next_config = _config[0]
            # Remove resource attribute
            config = self.configspace_ext.remove_resource(next_config)
            if self.do_profile:
                self.profiler.stop('all')
                self.profiler.pop_prefix()  # getconfig

        if self.debug_log is not None:
            self.debug_log.set_final_config(config)
            # All get_config debug log info is only written here
            self.debug_log.write_block()
        if self.do_profile:
            self.profiler.stop('all')
            self.profiler.clear()
        if self.getconfig_callback is not None:
            self.getconfig_callback(config.get_dictionary())

        return config

    def evaluation_failed(self, config: CS.Configuration):
        # Remove all pending evaluations for config
        self.cleanup_pending(config)
        # Mark config as failed (which means it will be blacklisted in
        # future get_config calls)
        # We need to create an extended config by appending a resource
        # attribute. Its value does not matter, because of how the blacklist
        # is created
        lowest_attr_value = self.configspace_ext.resource_attr_range[0]
        config_ext = self.configspace_ext.get(config, lowest_attr_value)
        self.state_transformer.mark_candidate_failed(config_ext)

    def cleanup_pending(self, config: CS.Configuration):
        """
        Removes all pending candidates whose configuration (i.e., lacking the
        resource attribute) is equal to config.
        This should be called after an evaluation terminates. For various
        reasons (e.g., termination due to convergence), pending candidates
        for this evaluation may still be present.
        It is also called for a failed evaluation.

        :param config: See above
        """
        config_dct = config.get_dictionary()

        def filter_pred(x: PendingEvaluation) -> bool:
            x_dct = self.configspace_ext.remove_resource(
                x.candidate, as_dict=True)
            return x_dct != config_dct

        self.state_transformer.filter_pending_evaluations(filter_pred)

    def remove_case(self, config: CS.Configuration, resource: int):
        config_ext = self.configspace_ext.get(config, resource)
        self.state_transformer.drop_candidate(config_ext)

    def clone_from_state(self, state):
        # Create clone with mutable state taken from 'state'
        model_args = self.state_transformer._model_args
        init_state = decode_state(
            state['state'], self.configspace_ext.hp_ranges_ext)
        skip_optimization = state['skip_optimization']
        new_searcher = GPMultiFidelitySearcher(
            hp_ranges=self.hp_ranges,
            resource_attr_key=self.configspace_ext.resource_attr_key,
            resource_attr_range=self.configspace_ext.resource_attr_range,
            random_seed=self.random_seed,
            gpmodel=self.state_transformer._gpmodel,
            model_args=model_args,
            map_reward=self.map_reward,
            acquisition_class=self.acquisition_class,
            resource_for_acquisition=self.resource_for_acquisition,
            init_state=init_state,
            local_minimizer_class=self.local_minimizer_class,
            skip_optimization=skip_optimization,
            num_initial_candidates=self.num_initial_candidates,
            num_initial_random_choices=self.num_initial_random_choices,
            initial_scoring=self.initial_scoring,
            profiler=self.profiler,
            first_is_default=self.first_is_default,
            debug_log=self.debug_log)
        new_searcher.state_transformer.set_params(state['model_params'])
        new_searcher.random_state = state['random_state']
        new_searcher.random_generator.random_state = \
            state['random_state']
        if self.debug_log and 'debug_log' in state:
            new_searcher.debug_log.set_mutable_state(state['debug_log'])
        # Invalidate self (must not be used afterwards)
        self.state_transformer = None
        return new_searcher


def draw_random_candidate(
        blacklisted_candidates: Set[CS.Configuration],
        configspace_ext: ExtendedConfiguration,
        random_state: np.random.RandomState,
        target_resource: int) -> (CS.Configuration, CS.Configuration):
    config, config_ext = None, None
    for _ in range(GET_CONFIG_RANDOM_RETRIES):
        _config = configspace_ext.hp_ranges.random_candidate(random_state)
        # Test whether this config has already been considered
        _config_ext = configspace_ext.remap_resource(
            _config, target_resource)
        if _config_ext not in blacklisted_candidates:
            config = _config
            config_ext = _config_ext
            break
    if config is None:
        raise AssertionError(
            "Failed to sample a configuration not already chosen "
            "before. Maybe there are no free configurations left? "
            "The blacklist size is {}".format(len(blacklisted_candidates)))
    return config, config_ext


def _max_at_least_k(k, counter=None):
    """
    Get largest key of `counter` whose value is at least `k`.

    :param counter: dict with keys that support comparison operators
    :param k: lower bound
    :return: largest key of `counter`
    """
    assert counter, "counter must be non-empty and not `None`."
    return max(filter(lambda r: counter[r] >= k, counter.keys()),
               default=min(counter.keys()))


def resource_for_acquisition_bohb(threshold: int) -> Callable[..., int]:
    """
    Factory for resource_for_acquisition argument in GPMultiFidelitySearcher.
    Mirrors what is done in the BOHB algorithm. An integer threshold is given
    at construction.
    We return the largest resource level whose number of labeled candidates
    is >= the threshold. If none of the levels passes the threshold, the
    lowest level is returned.

    """
    def rfa_map(state: TuningJobState, resource_attr_name: str, **kwargs) -> int:
        assert state.candidate_evaluations, "state must not be empty"
        histogram = Counter([int(x.candidate.get_dictionary()[resource_attr_name])
                             for x in state.candidate_evaluations])
        return _max_at_least_k(threshold, histogram)

    return rfa_map


def resource_for_acquisition_first_milestone(
        state: TuningJobState, resource_attr_name: str, **kwargs) -> int:
    """
    Implementation for resource_for_acquisition argument in
    GPMultiFidelitySearcher. We assume that the scheduler passes the first
    milestone to be attained by the new config as kwargs['milestone']. This is
    returned as resource level.

    """
    assert 'milestone' in kwargs, \
        "Need the first milestone to be attained by the new config passed as "\
        "kwargs['milestone']. Use a scheduler which does that (in particular, "\
        "Hyperband_Scheduler)"
    return kwargs['milestone']



import logging
from collections import defaultdict

import numpy as np
import pandas as pd
from pandas import DataFrame, Series

logger = logging.getLogger(__name__)


def get_type_family_raw(dtype) -> str:
    """From dtype, gets the dtype family."""
    try:
        if dtype.name == 'category':
            return 'category'
        if 'datetime' in dtype.name:
            return 'datetime'
        elif np.issubdtype(dtype, np.integer):
            return 'int'
        elif np.issubdtype(dtype, np.floating):
            return 'float'
    except Exception as err:
        logger.error(f'Warning: dtype {dtype} is not recognized as a valid dtype by numpy! AutoGluon may incorrectly handle this feature...')
        logger.error(err)

    if dtype.name in ['bool', 'bool_']:
        return 'bool'
    elif dtype.name in ['str', 'string', 'object']:
        return 'object'
    else:
        return dtype.name


# Real dtypes
def get_type_map_real(df: DataFrame) -> dict:
    features_types = df.dtypes.to_dict()
    return {k: v.name for k, v in features_types.items()}


# Raw dtypes (Real dtypes family)
def get_type_map_raw(df: DataFrame) -> dict:
    features_types = df.dtypes.to_dict()
    return {k: get_type_family_raw(v) for k, v in features_types.items()}


def get_type_map_special(X: DataFrame) -> dict:
    type_map_special = {}
    for column in X:
        type_special = get_type_special(X[column])
        if type_special is not None:
            type_map_special[column] = type_special
    return type_map_special


def get_type_special(X: Series) -> str:
    if check_if_datetime_as_object_feature(X):
        type_special = 'datetime_as_object'
    elif check_if_nlp_feature(X):
        type_special = 'text'
    else:
        type_special = None
    return type_special


def get_type_group_map(type_map: dict) -> defaultdict:
    type_group_map = defaultdict(list)
    for key, val in type_map.items():
        type_group_map[val].append(key)
    return type_group_map


def get_type_group_map_real(df: DataFrame) -> defaultdict:
    type_map_real = get_type_map_real(df)
    return get_type_group_map(type_map_real)


def get_type_group_map_raw(df: DataFrame) -> defaultdict:
    type_map_raw = get_type_map_raw(df)
    return get_type_group_map(type_map_raw)


# TODO: Expand to enable multiple special types per feature
def get_type_group_map_special(df: DataFrame) -> defaultdict:
    type_map_special = get_type_map_special(df)
    return get_type_group_map(type_map_special)


# TODO: Expand to int64 -> date features (milli from epoch etc)
# TODO: This takes a surprisingly long time to run, ~30 seconds a laptop for 50,000 rows of datetime_as_object for a single column. Try to optimize.
def check_if_datetime_as_object_feature(X: Series) -> bool:
    type_family = get_type_family_raw(X.dtype)
    # TODO: Check if low numeric numbers, could be categorical encoding!
    # TODO: If low numeric, potentially it is just numeric instead of date
    if X.isnull().all():
        return False
    if type_family != 'object':  # TODO: seconds from epoch support
        return False
    try:
        # TODO: pd.Series(['20170204','20170205','20170206']) is incorrectly not detected as datetime_as_object
        #  But we don't want pd.Series(['184','822828','20170206']) to be detected as datetime_as_object
        #  Need some smart logic (check min/max values?, check last 2 values don't go >31?)
        X.apply(pd.to_numeric)
    except:
        try:
            result = X.apply(pd.to_datetime, errors='coerce')
            if result.isnull().mean() > 0.8:  # If over 80% of the rows are NaN
                return False
            return True
        except:
            return False
    else:
        return False


def check_if_nlp_feature(X: Series) -> bool:
    type_family = get_type_family_raw(X.dtype)
    if type_family != 'object':
        return False
    X_unique = X.unique()
    num_unique = len(X_unique)
    num_rows = len(X)
    unique_ratio = num_unique / num_rows
    if unique_ratio <= 0.01:
        return False
    try:
        avg_words = Series(X_unique).str.split().str.len().mean()
    except AttributeError:
        return False
    if avg_words < 3:
        return False

    return True



"""Distributed Task Scheduler"""
import logging
import pickle
import sys
from collections import OrderedDict
from warnings import warn

import distributed

from .jobs import DistributedJobRunner
from .managers import TaskManagers
from .. import Task
from ..utils import AutoGluonWarning

logger = logging.getLogger(__name__)

__all__ = ['TaskScheduler']


class ClassProperty(object):

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class TaskScheduler(object):
    """Base Distributed Task Scheduler
    """
    managers = TaskManagers()
    jobs = DistributedJobRunner()

    def __init__(self, dist_ip_addrs=None):
        cls = TaskScheduler
        cls.managers.register_dist_ip_addrs(dist_ip_addrs)
        self.scheduled_tasks = []
        self.finished_tasks = []

    def add_remote(self, ip_addrs):
        """Add remote nodes to the scheduler computation resource.
        """
        self.managers.add_remote(ip_addrs)

    @classmethod
    def upload_files(cls, files, **kwargs):
        """Upload files to remote machines, so that they are accessible by import or load.
        """
        cls.managers.upload_files(files, **kwargs)

    def _dict_from_task(self, task):
        if isinstance(task, Task):
            return {'TASK_ID': task.task_id, 'Args': task.args}
        else:
            assert isinstance(task, dict)
            return {'TASK_ID': task['TASK_ID'], 'Args': task['Args']}

    def add_task(self, task, **kwargs):
        """add_task() is now deprecated in favor of add_job().
        """
        warn("scheduler.add_task() is now deprecated in favor of scheduler.add_job().",
             AutoGluonWarning)
        self.add_job(task, **kwargs)

    def add_job(self, task, **kwargs):
        """Adding a training task to the scheduler.

        Args:
            task (:class:`autogluon.scheduler.Task`): a new training task

        Relevant entries in kwargs:
        - bracket: HB bracket to be used. Has been sampled in _promote_config
        - new_config: If True, task starts new config eval, otherwise it promotes
          a config (only if type == 'promotion')
        Only if new_config == False:
        - config_key: Internal key for config
        - resume_from: config promoted from this milestone
        - milestone: config promoted to this milestone (next from resume_from)
        """
        # adding the task
        cls = TaskScheduler
        if not task.resources.is_ready:
            cls.managers.request_resources(task.resources)
        job = cls.jobs.start_distributed_job(task, cls.managers)
        new_dict = self._dict_from_task(task)
        new_dict['Job'] = job
        with self.managers.lock:
            self.scheduled_tasks.append(new_dict)

    def run_job(self, task):
        """Run a training task to the scheduler (Sync).
        """
        cls = TaskScheduler
        cls.managers.request_resources(task.resources)
        job = cls.jobs.start_distributed_job(task, cls.managers)
        return job.result()

    def _clean_task_internal(self, task_dict):
        pass

    def _cleaning_tasks(self):
        with self.managers.lock:
            new_scheduled_tasks = []
            for task_dict in self.scheduled_tasks:
                if task_dict['Job'].done():
                    self._clean_task_internal(task_dict)
                    self.finished_tasks.append(self._dict_from_task(task_dict))
                else:
                    new_scheduled_tasks.append(task_dict)
            if len(new_scheduled_tasks) < len(self.scheduled_tasks):
                self.scheduled_tasks = new_scheduled_tasks

    def join_tasks(self):
        warn("scheduler.join_tasks() is now deprecated in favor of scheduler.join_jobs().",
             AutoGluonWarning)
        self.join_jobs()

    def join_jobs(self, timeout=None):
        """Wait all scheduled jobs to finish
        """
        self._cleaning_tasks()
        for task_dict in self.scheduled_tasks:
            try:
                task_dict['Job'].result(timeout=timeout)
            except distributed.TimeoutError as e:
                logger.error(str(e))
            except:
                logger.error("Unexpected error:", sys.exc_info()[0])
                raise
            self._clean_task_internal(task_dict)
        self._cleaning_tasks()

    def shutdown(self):
        """shutdown() is now deprecated in favor of :func:`autogluon.done`.
        """
        warn("scheduler.shutdown() is now deprecated in favor of autogluon.done().",
             AutoGluonWarning)
        self.join_jobs()
        self.remote_manager.shutdown()

    def state_dict(self, destination=None):
        """Returns a dictionary containing a whole state of the Scheduler

        Examples
        --------
        >>> ag.save(scheduler.state_dict(), 'checkpoint.ag')
        """
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination['finished_tasks'] = pickle.dumps(self.finished_tasks)
        destination['TASK_ID'] = Task.TASK_ID.value
        return destination

    def load_state_dict(self, state_dict):
        """Load from the saved state dict.

        Examples
        --------
        >>> scheduler.load_state_dict(ag.load('checkpoint.ag'))
        """
        self.finished_tasks = pickle.loads(state_dict['finished_tasks'])
        Task.set_id(state_dict['TASK_ID'])
        logger.debug('\nLoading finished_tasks: {} '.format(self.finished_tasks))

    @property
    def num_finished_tasks(self):
        return len(self.finished_tasks)

    def __repr__(self):
        reprstr = self.__class__.__name__ + '(\n' + \
                  str(self.managers.resource_manager) + ')\n'
        return reprstr



import autograd.numpy as anp

from .base import KernelFunction
from ..constants import COVARIANCE_SCALE_LOWER_BOUND, \
    COVARIANCE_SCALE_UPPER_BOUND, DEFAULT_ENCODING
from ..gluon_blocks_helpers import encode_unwrap_parameter, IdentityScalarEncoding
from ..utils import create_encoding, register_parameter

__all__ = ['FabolasKernelFunction']


class FabolasKernelFunction(KernelFunction):
    """
    The kernel function proposed in:

        Klein, A., Falkner, S., Bartels, S., Hennig, P., & Hutter, np. (2016).
        Fast Bayesian Optimization of Machine Learning Hyperparameters
        on Large Datasets, in AISTATS 2017.
        ArXiv:1605.07079 [Cs, Stat]. Retrieved from http://arxiv.org/abs/1605.07079

    Please note this is only one of the components of the factorized kernel
    proposed in the paper. This is the finite-rank ("degenerate") kernel for
    modelling data subset fraction sizes. Defined as:

        k(x, y) = (U phi(x))^T (U phi(y)),  x, y in [0, 1],
        phi(x) = [1, (1 - x)^2]^T,  U = [[u1, u3], [0, u2]] upper triangular,
        u1, u2 > 0.
    """
    def __init__(self, dimension=1, encoding_type=DEFAULT_ENCODING,
                 u1_init=1.0, u3_init=0.0, **kwargs):
        super(FabolasKernelFunction, self).__init__(dimension=dimension, **kwargs)
        self.encoding_u12 = create_encoding(
            encoding_type, u1_init, COVARIANCE_SCALE_LOWER_BOUND,
            COVARIANCE_SCALE_UPPER_BOUND, 1, None)
        # This is not really needed, but param_encoding_pairs needs an encoding
        # for each parameter
        self.encoding_u3 = IdentityScalarEncoding(init_val=u3_init)
        with self.name_scope():
            self.u1_internal = register_parameter(
                self.params, 'u1', self.encoding_u12)
            self.u2_internal = register_parameter(
                self.params, 'u2', self.encoding_u12)
            self.u3_internal = register_parameter(
                self.params, 'u3', self.encoding_u3)

    @staticmethod
    def _compute_factor(x, u1, u2, u3):
        tvec = (1.0 - x) ** 2
        return anp.concatenate(
           [anp.add(anp.multiply(tvec, u3), u1),
            anp.multiply(tvec, u2)], axis=1)

    def forward(self, X1, X2):
        u1_internal = self.u1_internal.data()
        u2_internal = self.u1_internal.data()
        u3_internal = self.u1_internal.data()
        X1 = self._check_input_shape(X1)
        u1 = self.encoding_u12.get(u1_internal)
        u2 = self.encoding_u12.get(u2_internal)
        u3 = self.encoding_u3.get(u3_internal)
        mat1 = self._compute_factor(X1, u1, u2, u3)
        if X2 is X1:
            return anp.dot(mat1, anp.transpose(mat1))
        else:
            X2 = self._check_input_shape(X2)
            mat2 = self._compute_factor(X2, u1, u2, u3)
            return anp.dot(mat1, anp.transpose(mat2))

    def _get_pars(self, X):
        u1 = encode_unwrap_parameter(self.u1_internal, self.encoding_u12, X)
        u2 = encode_unwrap_parameter(self.u2_internal, self.encoding_u12, X)
        u3 = encode_unwrap_parameter(self.u3_internal, self.encoding_u3, X)
        return (u1, u2, u3)

    def diagonal(self, X):
        X = self._check_input_shape(X)
        u1, u2, u3 = self._get_pars(X)
        mat = self._compute_factor(X, u1, u2, u3)
        return anp.sum(mat ** 2, axis=1)

    def diagonal_depends_on_X(self):
        return True

    def param_encoding_pairs(self):
        return [
            (self.u1_internal, self.encoding_u12),
            (self.u2_internal, self.encoding_u12),
            (self.u3_internal, self.encoding_u3)
        ]

    def get_params(self):
        values = list(self._get_pars(None))
        keys = ['u1', 'u2', 'u3']
        return {k: anp.reshape(v, (1,))[0] for k, v in zip(keys, values)}

    def set_params(self, param_dict):
        self.encoding_u12.set(self.u1_internal, param_dict['u1'])
        self.encoding_u12.set(self.u2_internal, param_dict['u2'])
        self.encoding_u3.set(self.u3_internal, param_dict['u3'])
